# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.internal.mailbox import InternalMailbox
from .mock_compat import Mock


@pytest.fixture
def actor():
    actor = Mock()
    actor.handle_message = Mock()
    actor.handle_system_message = Mock()
    return actor


@pytest.fixture
def mailbox(actor):
    return InternalMailbox(Mock(), actor)


def test_should_invoke_handle_message_on_actor(actor, mailbox):
    message = object()
    mailbox.enqueue(message)
    mailbox.process_messages()
    actor.handle_message.assert_called_once_with(message)


def test_should_invoke_handle_system_message_on_actor(actor, mailbox):
    message = object()
    mailbox.enqueue_system(message)
    mailbox.process_messages()
    actor.handle_system_message.assert_called_once_with(message)


def test_should_process_system_messages_before_user_messages(actor, mailbox):
    mailbox.enqueue(1)
    mailbox.enqueue_system(2)

    def assert_handle_message_not_called(*args, **kwargs):
        assert not actor.handle_message.called
    actor.handle_system_message.side_effect = assert_handle_message_not_called

    mailbox.process_messages()
    actor.handle_system_message.assert_called_once_with(2)
    actor.handle_message.assert_called_once_with(1)


def test_should_stop_processing_when_reaching_throughput_limit(actor):
    mailbox = InternalMailbox(Mock(), actor, throughput=2)
    mailbox.enqueue(1)
    mailbox.enqueue(1)
    mailbox.enqueue(1)
    mailbox.process_messages()
    assert actor.handle_message.call_count == 2


def test_should_not_processing_user_messages_when_suspended(actor, mailbox):
    mailbox.suspend()
    mailbox.enqueue(Mock())
    mailbox.process_messages()
    assert not actor.handle_system_message.called


def test_should_process_system_messages_when_suspended(actor, mailbox):
    mailbox.suspend()
    mailbox.enqueue_system(Mock())
    mailbox.process_messages()
    assert actor.handle_system_message.called


def test_should_not_process_messages_when_closed(actor, mailbox):
    mailbox.close()
    mailbox.enqueue(Mock())
    mailbox.enqueue_system(Mock())
    mailbox.process_messages()
    assert not actor.handle_message.called
    assert not actor.handle_system_message.called


def test_should_process_user_messages_on_resume(actor, mailbox):
    mailbox.suspend()
    mailbox.enqueue(Mock())
    mailbox.resume()
    mailbox.process_messages()
    assert actor.handle_message.called


def test_suspending_in_handler_interrupts_processing(actor):
    mailbox = InternalMailbox(Mock(), actor, throughput=5)
    actor.handle_message.side_effect = lambda *args: mailbox.suspend()
    mailbox.enqueue(Mock())
    mailbox.enqueue(Mock())
    mailbox.process_messages()
    assert actor.handle_message.call_count == 1


def test_closing_in_handler_interrupts_processing(actor):
    mailbox = InternalMailbox(Mock(), actor, throughput=5)
    actor.handle_system_message.side_effect = lambda *args: mailbox.close()
    mailbox.enqueue_system(Mock())
    mailbox.enqueue_system(Mock())
    mailbox.enqueue(Mock())
    mailbox.enqueue(Mock())
    mailbox.process_messages()
    assert actor.handle_system_message.call_count == 1


def test_set_scheduled_while_already_scheduled_should_fail(mailbox):
    assert mailbox.set_scheduled() is True
    assert mailbox.is_scheduled()
    assert mailbox.set_scheduled() is False


def test_sets_itself_idle_after_processing_messages(mailbox):
    mailbox.enqueue(Mock())
    mailbox.set_scheduled()
    mailbox.process_messages()
    assert mailbox.is_idle()


def test_flushing_when_open_should_raise(mailbox):
    mailbox.enqueue(Mock())
    with pytest.raises(Exception):
        mailbox.flush_messages()


def test_flushing_when_closed_should_return_user_messages_in_queue(mailbox):
    message = object()
    mailbox.enqueue(message)
    mailbox.close()
    assert mailbox.flush_messages() == [message]




# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import actors.internal.cell
import actors.internal.factory
from actors.future import Promise
from actors.ref import ActorRef, InternalRef
from actors.actor import Actor
from actors.internal.dispatcher import Dispatcher
from actors.internal.messages import Terminate, Start, DeadLetter
from actors.internal.executor import Executor


class ActorSystem(actors.internal.factory.ActorFactory):

    def __init__(self, system_dispatcher=None):
        """
        The actor system is responsible for creating, configuring and stopping actors and
        dispatchers.

        Normally, only one system per application should be created.

        :param system_dispatcher: Override the dispatcher used by the system. This also acts as the
            default dispatcher for new actors.
        :type system_dispatcher: :class:`Dispatcher`
        """
        self._system_dispatcher = Dispatcher(Executor()) \
            if system_dispatcher is None else system_dispatcher
        self._dead_letters = _DeadLetterRef()

        self._terminate_promise = Promise()

        class Guardian(Actor):
            def __init__(me):
                me._logger = logging.getLogger(__name__)

            def receive(me, message):
                me._logger.warning("User receive called. This should not be happen.")

            def post_stop(me):
                self._terminate_promise.complete(None)

        self._guardian = InternalRef(actors.internal.cell.Cell(Guardian,
            dispatcher=self._system_dispatcher, system=self, parent=None))
        self._guardian.send_system_message(Start)

        actors.internal.factory.ActorFactory.__init__(self, self, self._guardian)

    def terminate(self):
        self._guardian.send_system_message(Terminate)
        self._terminate_promise.future.get()

    @property
    def dead_letters(self):
        return self._dead_letters

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()


class _DeadLetterRef(ActorRef):
    def __init__(self):
        super(_DeadLetterRef, self).__init__(None)
        self._logger = logging.getLogger('dead letter')

    def tell(self, message, sender=None):
        if isinstance(message, DeadLetter):
            self._logger.debug("Message %r from %r to %r was not delivered.",
                message.message, message.sender, message.recipient)
        else:
            self._logger.debug("Message %r from %r was not delivered.")



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from threading import Thread
from six.moves.queue import Queue


logger = logging.getLogger(__name__)


class Executor(object):
    _INTERRUPT = object()

    def __init__(self, num_workers=1):
        super(Executor, self).__init__()
        self._queue = Queue()
        self._workers = []

        for _ in range(num_workers):
            th = Thread(target=self._work)
            th.start()
            self._workers.append(th)

    def submit(self, task):
        self._queue.put(task)

    def shutdown(self):
        for _ in self._workers:
            self._queue.put(self._INTERRUPT)

    def join(self):
        for worker in self._workers:
            worker.join()

    def _work(self):
        while True:
            task = self._queue.get(block=True)
            if task is self._INTERRUPT:
                break
            try:
                task()
            except BaseException as e:
                logger.exception(e)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from actors.actor import Actor


class AsyncCountDownLatch(Actor):
    CountDown = object()
    Done = object()
    Start = namedtuple('Start', ['count'])

    def __init__(self):
        self._count = -1
        self._reply_to = None

    def receive(self, message):
        if message is AsyncCountDownLatch.CountDown:
            self._count -= 1
            if self._count == 0:
                self._reply_to.tell(AsyncCountDownLatch.Done)
        elif type(message) is AsyncCountDownLatch.Start:
            self._count = message.count
            self._reply_to = self.context.sender



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors import Actor, ActorSystem, Envelope
from actors.internal.cell import Cell
from actors.internal.dispatcher import Dispatcher
from actors.internal.mailbox import InternalMailbox
from actors.internal.messages import Start, Terminate, Supervise, DeadLetter, Terminated
from actors.ref import InternalRef, ActorRef
from .mock_compat import Mock, MagicMock


@pytest.fixture
def actor():
    return Mock(spec_set=Actor)


@pytest.fixture
def supervisor():
    return Mock(spec_set=Cell)


@pytest.fixture
def dispatcher():
    return Mock(spec_set=Dispatcher)


@pytest.fixture
def mailbox():
    return MagicMock(spec_set=InternalMailbox)


@pytest.fixture
def system():
    return MagicMock(spec_set=ActorSystem)


@pytest.fixture
def cell(actor, system, dispatcher, supervisor, mailbox):
    cell = Cell(lambda: actor, system, dispatcher, supervisor)
    cell._mailbox = mailbox
    cell.handle_system_message(Start)
    return cell


def test_terminate_calls_stop_hook(actor, cell):
    cell.handle_system_message(Terminate)
    actor.post_stop.assert_called_once_with()


def test_terminate_closes_mailbox(cell, mailbox):
    cell.handle_system_message(Terminate)
    mailbox.close.assert_called_once_with()


def test_terminate_flush_mailbox(cell, mailbox):
    cell.handle_system_message(Terminate)
    mailbox.flush_messages.assert_called_once_with()


def test_terminate_sends_unhandled_messages_to_dead_letters(cell, mailbox, system):
    message = Mock()
    sender = Mock()
    mailbox.flush_messages.return_value = [Envelope(message, sender)]
    cell.handle_system_message(Terminate)
    system.dead_letters.tell.assert_called_once_with(DeadLetter(message, sender, ActorRef(cell)))


def test_sending_message_to_terminated_actor_should_forward_to_dead_letters(cell, system):
    cell.handle_system_message(Terminate)
    message = Mock()
    sender = Mock()
    cell.send_message(message, sender)
    system.dead_letters.tell.assert_called_once_with(DeadLetter(message, sender, ActorRef(cell)))


def test_terminate_detach_actor_from_dispatcher(cell, mailbox, dispatcher):
    cell.handle_system_message(Terminate)
    dispatcher.detach.assert_called_once_with(mailbox)


def test_terminate_sends_terminate_message_to_children(cell):
    child = Mock(spec_set=InternalRef)
    cell.handle_system_message(Supervise(child))
    cell.handle_system_message(Terminate)
    child.send_system_message.assert_called_once_with(Terminate)


def test_terminate_sends_terminated_message_to_supervisor(cell, supervisor):
    cell.handle_system_message(Terminate)
    supervisor.send_system_message.assert_called_once_with(Terminated(InternalRef(cell)))


def test_supervising_actors_finalize_after_children_have_terminated(actor, cell, mailbox,
        supervisor, dispatcher):
    child = Mock(spec_set=InternalRef)
    cell.handle_system_message(Supervise(child))

    cell.handle_system_message(Terminate)
    assert not dispatcher.detach.called
    assert not supervisor.send_system_message.called
    assert not actor.post_stop.called

    cell.handle_system_message(Terminated(child))
    dispatcher.detach.assert_called_once_with(mailbox)
    supervisor.send_system_message.assert_called_once_with(Terminated(InternalRef(cell)))
    actor.post_stop.assert_called_once_with()


def test_terminates_suspends_mailbox___closes_after_children_have_terminated(cell, mailbox):
    child = Mock(spec_set=InternalRef)
    cell.handle_system_message(Supervise(child))

    cell.handle_system_message(Terminate)
    mailbox.suspend.assert_called_once_with()
    assert not mailbox.close.called

    cell.handle_system_message(Terminated(child))
    mailbox.close.assert_called_once_with()


def test_terminating_children_does_not_terminate_supervisor(actor, cell, mailbox, supervisor, dispatcher):
    child = Mock(spec_set=InternalRef)
    cell.handle_system_message(Supervise(child))

    cell.handle_system_message(Terminated(child))
    assert not dispatcher.detach.called
    assert not supervisor.send_system_message.called
    assert not actor.post_stop.called
    assert not mailbox.suspend.called
    assert not mailbox.close.called




# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from actors.future import Promise
from actors.ref import ActorRef


class PromiseActorRef(ActorRef):
    def __init__(self):
        super(PromiseActorRef, self).__init__(None)
        self.promise = Promise()

    def tell(self, message, sender=None):
        self.promise.complete(message)


def ask(actor, message):
    """
    Send a message to `actor` and return a :class:`Future` holding a possible
    reply.

    To receive a result, the actor MUST send a reply to `sender`.

    :param actor:
    :type actor: :class:`ActorRef`.

    :param message:
    :type message: :type: Any

    :return: A future holding the result.
    """
    sender = PromiseActorRef()
    actor.tell(message, sender)
    return sender.promise.future




# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import no_sender


class ActorRef(object):
    __slots__ = ('_cell',)

    def __init__(self, _cell):
        self._cell = _cell

    def tell(self, message, sender=no_sender):
        """ Send a message to this actor. Asynchronous fire-and-forget.

        :param message: The message to send.
        :type message: Any

        :param sender: The sender of the message. If provided it will be made
            available to the receiving actor via the :attr:`Actor.sender` attribute.
        :type sender: :class:`Actor`
        """
        if sender is not no_sender and not isinstance(sender, ActorRef):
            raise ValueError("Sender must be actor reference")

        self._cell.send_message(message, sender)

    def __eq__(self, other):
        return isinstance(other, ActorRef) and other._cell is self._cell


class InternalRef(ActorRef):

    def __init__(self, cell):
        super(InternalRef, self).__init__(cell)

    def send_system_message(self, message):
        self._cell.send_system_message(message)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from threading import Lock


class _Status(object):
    OPEN = object()
    CLOSED = object()
    SUSPENDED = object()


class _ScheduleState(object):
    IDLE = 1
    SCHEDULED = 2


class InternalMailbox(object):
    def __init__(self, dispatcher, actor, throughput=5):
        # Should only be set in message handler
        self._primary_status = _Status.OPEN

        self._idle = _ScheduleState.IDLE
        self._idle_lock = Lock()

        self._message_queue = deque()
        self._system_message_queue = deque()
        self._actor = actor
        self._dispatcher = dispatcher
        self._throughput = throughput

    def enqueue(self, item):
        self._message_queue.append(item)

    def enqueue_system(self, item):
        self._system_message_queue.append(item)

    def close(self):
        """ Stop processing of all messages."""
        self._primary_status = _Status.CLOSED

    def suspend(self):
        """ Stop processing of user message. """
        self._primary_status = _Status.SUSPENDED

    def resume(self):
        """ Continue processing of user message. """
        self._primary_status = _Status.OPEN

    def is_closed(self):
        return self._primary_status is _Status.CLOSED

    def is_suspended(self):
        return self._primary_status is _Status.SUSPENDED

    def has_messages(self):
        return len(self._message_queue) > 0 or len(self._system_message_queue) > 0

    def is_idle(self):
        return self._idle

    def is_scheduled(self):
        return not self._idle

    def set_scheduled(self):
        """
        Returns True if state was successfully changed from idle to scheduled.
        """
        with self._idle_lock:
            if self._idle:
                self._idle = False
                return True
        return False

    def set_idle(self):
        with self._idle_lock:
            self._idle = True

    def process_messages(self):
        if self._primary_status is _Status.CLOSED or not self.has_messages():
            return

        self._process_system_message()

        if self._primary_status is _Status.OPEN:
            for _ in range(min(self._throughput, len(self._message_queue))):
                self._actor.handle_message(self._message_queue.popleft())
                self._process_system_message()
                if self._primary_status is not _Status.OPEN:
                    break

        self.set_idle()
        self._dispatcher.schedule_execution(self)

    def _process_system_message(self):
        if len(self._system_message_queue) > 0:
            self._actor.handle_system_message(self._system_message_queue.popleft())

    def flush_messages(self):
        if not self.is_closed():
            raise Exception()

        messages = []
        while len(self._message_queue) > 0:
            messages.append(self._message_queue.popleft())
        return messages



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import default_supervisor_strategy


class Actor(object):
    """ The base class implemented by all actors.

    :ivar context: Provides contextual information about this actor and the
        current message. Only valid within this actor.
    :vartype context: :class:`ActorContext`

    """

    context = None

    supervisor_strategy = default_supervisor_strategy
    """ Class method. Override to provide custom supervisor behaviour for this actor. """

    def receive(self, message):
        """
        Override to provide the actor behaviour.

        :param message: The current message.
        """

    def pre_start(self):
        """ Called asynchronous before processing any messages. """

    def post_stop(self):
        """ Called asynchronous after the actor has stopped. """

    def pre_restart(self):
        """ Called on the failed actor before it's disposed of. """

    def post_restart(self):
        """ Called on the newly created actor. """

    @property
    def sender(self):
        """ Alias for :attr:`self.context.sender`. """
        return self.context.sender

    @property
    def ref(self):
        """ Alias for :attr:`self.context.self_ref`. """
        return self.context.self_ref



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple


__all__ = [
    'no_sender',
    'Actor',
    'ActorContext',
    'ActorSystem',
    'ActorRef',
    'Directive',
]

from actors.internal import messages

Envelope = namedtuple('Envelope', ['message', 'sender'])

no_sender = None


class ActorInitializationError(Exception):
    pass


class Directive(object):
    Stop = messages.Terminate
    Restart = messages.Restart
    Resume = messages.Resume


def default_supervisor_strategy(exception):
    if isinstance(exception, ActorInitializationError):
        return Directive.Stop
    return Directive.Restart


from .actor import Actor
from .context import ActorContext
from .system import ActorSystem
from .ref import ActorRef



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.actor import Actor
from actors.internal.cell import Cell
from actors.internal.messages import Restart, Start
from .mock_compat import Mock


@pytest.fixture
def actor():
    return Mock(spec_set=Actor)


@pytest.fixture
def factory(actor):
    return Mock(return_value=actor)


@pytest.fixture
def cell(factory):
    cell = Cell(factory, Mock(), Mock(), Mock())
    cell.handle_system_message(Start)
    return cell


def test_restart_constructs_new_instance(factory, cell):
    assert factory.call_count == 1
    cell.handle_system_message(Restart)
    assert factory.call_count == 2


def test_restart_calls_pre_restart_on_old_and_post_on_new():
    actor1 = Mock(spec_set=Actor)
    actor2 = Mock(spec_set=Actor)

    it = iter([actor1, actor2])

    cell = Cell(lambda: next(it), Mock(), Mock(), Mock())
    cell.handle_system_message(Start)

    cell.handle_system_message(Restart)
    actor1.pre_restart.assert_called_once_with()
    actor2.post_restart.assert_called_once_with()


def test_failure_in_restart_hooks_should_stop_actor():
    pass



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from actors import ActorRef
from actors.utils.proxy import ProxyActor, ProxyRef, Invoke
from .mock_compat import Mock


def test_invoking_method_on_ref_sends_invoke_message():
    underlying = Mock(spec_set=ActorRef)
    ref = ProxyRef(underlying)
    ref.foo(1, 2, bar=3)
    underlying.tell.assert_called_once_with(Invoke('foo', (1, 2), {'bar': 3}))


def test_sending_invoke_message_calls_method():
    actor = ProxyActor()
    actor.foo = Mock()
    actor.receive(Invoke('foo', (1, 2), {'bar': 3}))
    actor.foo.assert_called_once_with(1, 2, bar=3)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.internal.dispatcher import Dispatcher
from actors.internal.executor import Executor
from actors.internal.mailbox import InternalMailbox
from .mock_compat import Mock


@pytest.fixture
def executor():
    return Mock(spec_set=Executor)


@pytest.fixture
def dispatcher(executor):
    return Dispatcher(executor)


def test_dispatch_should_enqueue_message(dispatcher):
    mailbox = Mock(spec_set=InternalMailbox)
    mailbox.is_closed.return_value = False

    message = object()
    dispatcher.dispatch(message, mailbox)
    mailbox.enqueue.assert_called_once_with(message)


def test_dispatch_should_execute_mailbox(executor, dispatcher):
    mailbox = Mock(spec_set=InternalMailbox)
    mailbox.has_messages.return_value = True
    mailbox.is_closed.return_value = False
    mailbox.is_scheduled.return_value = False

    dispatcher.dispatch(Mock(), mailbox)
    executor.submit.assert_called_once_with(mailbox.process_messages)


def test_dispatch_should_not_execute_on_unsuccessful_schedule(executor, dispatcher):
    mailbox = Mock(spec_set=InternalMailbox)
    mailbox.has_messages.return_value = True
    mailbox.is_closed.return_value = False
    mailbox.is_scheduled.return_value = False
    mailbox.set_scheduled.return_value = False

    dispatcher.dispatch(Mock(), mailbox)
    assert mailbox.set_scheduled.called
    assert not executor.submit.called


@pytest.mark.xfail
def test_should_shutdown_executor_when_no_actors_attached(executor, dispatcher):
    dispatcher.await_shutdown()
    executor.shutdown.assert_called_once_with()


def test_dispatch_should_not_deliver_when_mailbox_is_closed(executor, dispatcher):
    mailbox = Mock(spec_set=InternalMailbox)
    mailbox.has_messages.return_value = True
    mailbox.is_closed.return_value = True

    dispatcher.dispatch(Mock(), mailbox)
    dispatcher.dispatch_system(Mock(), mailbox)

    assert not mailbox.enqueue.called
    assert not mailbox.set_scheduled.called
    assert not executor.submit.called



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.actor import Actor
from actors.ref import ActorRef
from .mock_compat import Mock
from actors import Envelope, ActorInitializationError
from actors.internal.cell import Cell
from actors.internal.messages import Start, Failure, Terminate, Restart, Terminated


@pytest.fixture
def actor():
    actor = Mock(spec_set=Actor)
    actor.receive = Mock()
    actor.pre_start = Mock()
    actor.post_stop = Mock()
    return actor


@pytest.fixture
def supervisor():
    supervisor = Mock()
    supervisor.send_system_message = Mock()
    return supervisor


@pytest.fixture
def running_cell(actor, supervisor):
    cell = Cell(lambda: actor, Mock(), Mock(), supervisor)
    cell.handle_system_message(Start)
    return cell


def test_handle_message_invokes_receive_on_actor(actor, running_cell):
    message = object()
    running_cell.handle_message(Envelope(message, None))
    actor.receive.assert_called_once_with(message)


def test_handle_message_should_raise_if_actor_not_created(actor):
    cell = Cell(lambda: actor, Mock(), Mock(), Mock())
    with pytest.raises(Exception):
        cell.handle_message(Mock())


def test_start_constructs_instance():
    factory = Mock()
    cell = Cell(factory, Mock(), Mock(), Mock())
    cell.handle_system_message(Start)
    factory.assert_called_once_with()


def test_start_calls_start_hook(actor):
    cell = Cell(lambda: actor, Mock(), Mock(), Mock())
    cell.handle_system_message(Start)
    actor.pre_start.assert_called_once_with()


# Supervision

def _check_sent_failure_message(expected_sender, receiver, expected_exception):
    message = receiver.send_system_message.call_args[0][0]
    assert type(message) is Failure
    assert message.ref == expected_sender
    assert type(message.exception) is type(expected_exception)


def test_failure_in_message_handler_should_suspend_mailbox(actor, running_cell):
    actor.receive.side_effect = ValueError()
    running_cell.handle_message(Mock())
    assert running_cell._mailbox.is_suspended()


def test_failure_in_message_handler_sends_failure_message_to_supervisor(actor, supervisor, running_cell):
    actor.receive.side_effect = ArithmeticError()
    running_cell.handle_message(Mock())
    _check_sent_failure_message(ActorRef(running_cell), supervisor, ArithmeticError())


def test_failure_in_start_hook_should_send_failure_to_supervisor(actor, supervisor):
    actor.pre_start.side_effect = AttributeError()
    cell = Cell(lambda: actor, Mock(), Mock(), supervisor)
    cell.handle_system_message(Start)
    actor.pre_start.assert_called_once_with()
    _check_sent_failure_message(ActorRef(cell), supervisor, ActorInitializationError())


def test_sends_failure_if_factory_does_not_return_an_actor(supervisor):
    cell = Cell(lambda: object(), Mock(), Mock(), supervisor)
    cell.handle_system_message(Start)
    _check_sent_failure_message(ActorRef(cell), supervisor, ActorInitializationError())


def test_failure_in_restart_hook_should_send_failure_to_supervisor(actor, supervisor, running_cell):
    actor.pre_restart.side_effect = ArithmeticError()
    running_cell.handle_system_message(Restart)
    _check_sent_failure_message(ActorRef(running_cell), supervisor, ActorInitializationError())


def test_failure_in_stop_hook_should_not_send_failure_to_supervisor(actor, supervisor, running_cell):
    actor.post_stop.side_effect = AttributeError()
    running_cell.handle_system_message(Terminate)

    actor.post_stop.assert_called_once_with()
    for call in supervisor.send_system_message.call_args_list:
        assert not isinstance(call[0][0], Failure)


def test_failure_in_start_hook_stops_actor(actor, supervisor, running_cell):
    pass


def test_failure_in_restart_hook_stops_actor(actor, supervisor, running_cell):
    pass


def test_context_properties(actor, supervisor, running_cell):
    sender = object()

    global actual_sender, actual_supervisor, actual_self_ref
    actual_sender = actual_supervisor = actual_self_ref = None

    def store_context(message):
        global actual_sender, actual_supervisor, actual_self_ref
        actual_sender = actor.context.sender
        actual_supervisor = actor.context.supervisor
        actual_self_ref = actor.context.self_ref

    actor.receive.side_effect = store_context

    running_cell.handle_message(Envelope(Mock(), sender))
    assert actor.receive.call_count == 1

    assert actual_sender is sender
    assert actual_supervisor is supervisor
    assert actual_self_ref == ActorRef(running_cell)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import traceback
import pytest
from actors.utils.ask import Promise
from .mock_compat import Mock


def test_get_should_return_on_success():
    promise = Promise()
    promise.success(1)
    assert promise.future.get() == 1


def test_get_should_raise_on_failure():
    promise = Promise()
    try:
        1/0
    except ZeroDivisionError:
        traceback.print_exc()
        promise.failure(sys.exc_info())

    with pytest.raises(ZeroDivisionError):
        promise.future.get()


def test_calls_callback_on_success():
    callback = Mock()
    promise = Promise()
    promise.future.on_success(callback)
    promise.success(None)
    callback.assert_called_once_with(promise.future)


def test_calls_callback_on_failure():
    callback = Mock()
    promise = Promise()
    promise.future.on_failure(callback)
    promise.failure(None)
    callback.assert_called_once_with(promise.future)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.internal.cell import Cell
from actors.ref import ActorRef, InternalRef
from .mock_compat import Mock


def test_tell_sends_message_and_sender_to_cell():
    cell = Mock(spec_set=Cell)
    sender = Mock(spec_set=ActorRef)
    message = object()
    ref = ActorRef(cell)
    ref.tell(message, sender)

    cell.send_message.assert_called_once_with(message, sender)


def test_tell_raises_if_sender_is_not_an_actor_ref():
    ref = ActorRef(Mock())
    with pytest.raises(ValueError):
        ref.tell(Mock(), object())


def test_equality():
    cell_1 = object()
    cell_2 = object()
    assert ActorRef(cell_1) != ActorRef(cell_2)
    assert ActorRef(cell_1) == ActorRef(cell_1)
    assert InternalRef(cell_1) != InternalRef(cell_2)
    assert InternalRef(cell_1) == InternalRef(cell_1)
    assert InternalRef(cell_1) == ActorRef(cell_1)
    assert ActorRef(cell_1) != object()





# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback

import actors.actor
import actors.context
import actors.system
from actors import ActorInitializationError, Envelope
from actors.actor import Actor
from actors.ref import ActorRef, InternalRef
from actors.internal.messages import Start, Terminate, Failure, Supervise, Terminated, Restart, \
    DeadLetter
from actors.internal.mailbox import InternalMailbox

logger = logging.getLogger(__name__)


class Cell(object):

    def __init__(self, actor_factory, system, dispatcher, parent):
        self._actor_factory = actor_factory  # Creates actor instances
        self._actor = None
        self._system = system
        self._dispatcher = dispatcher
        self._mailbox = InternalMailbox(dispatcher, self, dispatcher.throughput)
        self._dispatcher.attach(self._mailbox)

        self._parent = parent
        self._children = []
        self._stopped = False
        self._self_ref = InternalRef(self)

        self._context = actors.context.ActorContext(system, self._self_ref)
        self._context.system = system

        # TODO: shouldn't expose the internal ref
        self._context.self_ref = self._self_ref

    def send_message(self, message, sender):
        if not self._mailbox.is_closed():
            self._dispatcher.dispatch(Envelope(message, sender), self._mailbox)
        else:
            self._system.dead_letters.tell(DeadLetter(message, sender, self._self_ref))

    def send_system_message(self, message):
        if not self._mailbox.is_closed():
            self._dispatcher.dispatch_system(message, self._mailbox)

    def handle_message(self, envelope):
        assert self._actor is not None
        try:
            self._context.sender = envelope.sender
            self._context.supervisor = self._parent
            self._actor.context = self._context
            self._actor.receive(envelope.message)
        except BaseException as e:
            self._on_failure(e)
        self._actor.context = None

    def handle_system_message(self, message):
        if message is Start:
            try:
                assert self._actor is None
                self._actor = self._actor_factory()
                assert isinstance(self._actor, Actor)
                self._actor.pre_start()
                # logger.debug("created actor %r" % self)
            except BaseException as e:
                self._on_failure(ActorInitializationError())

        # Directives from supervisor
        elif message is Restart:
            try:
                self._actor.pre_restart()
                self._actor = self._actor_factory()
                assert isinstance(self._actor, Actor)
                self._actor.post_restart()
                self._mailbox.resume()
                logger.debug("Actor %s restarted" % self)
            except BaseException as e:
                # TODO: use a different exception
                self._on_failure(ActorInitializationError())

        elif message is Terminate:
            self._stopped = True
            self._mailbox.suspend()
            for child in self._children:
                child.send_system_message(Terminate)
            if len(self._children) == 0:
                self._shutdown()

        elif type(message) is Supervise:
            try:
                assert isinstance(message.ref, InternalRef)
                self._children.append(message.ref)
            except (AttributeError, AssertionError) as e:
                logger.error("Failed to remove child", exc_info=True)

        elif type(message) is Terminated:
            try:
                assert isinstance(message.ref, InternalRef)
                assert message.ref in self._children
                self._children.remove(message.ref)
            except (KeyError, AttributeError, AssertionError):
                logger.error("Failed to remove child", exc_info=True)

            if self._stopped and len(self._children) == 0:
                self._shutdown()

        elif type(message) is Failure:
            try:
                directive = self._actor.__class__.supervisor_strategy(message.exception)
                assert directive is not None
            except BaseException:
                logger.error("Failed to get directive.")
            else:
                logger.info("Failure in child %r, exception: %s, directive: %r",
                            message.ref, type(message.exception).__name__, directive)
                message.ref.send_system_message(directive)
        else:
            logger.error("Got unknown system message '%r'", message)

    def _shutdown(self):
        self._mailbox.close()
        self._dispatcher.detach(self._mailbox)
        try:
            self._actor.post_stop()
        except BaseException:
            logger.warn("Failure in post_stop", exc_info=True)
        self._actor = None

        for envelope in self._mailbox.flush_messages():
            self._system.dead_letters.tell(DeadLetter(
                envelope.message, envelope.sender, self._self_ref))

        if self._parent:
            self._parent.send_system_message(Terminated(self._self_ref))

    def _on_failure(self, exception):
        logger.debug("Failure in actor '%s'. traceback:\n%s", self._self_ref,
             traceback.format_exc())

        self._mailbox.suspend()
        if self._parent:
            msg = Failure(self._self_ref, exception, traceback.format_exc())
            self._parent.send_system_message(msg)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.internal.factory import ActorFactory
from actors.internal.messages import Supervise
from actors.ref import InternalRef
from tests.mock_compat import Mock


def test_actor_of_send_supervise_message_to_parent_on_creation():
    supervisor = Mock(spec_set=InternalRef)
    factory = ActorFactory(Mock(), supervisor)
    factory.actor_of(Mock())
    assert supervisor.send_system_message.called
    assert type(supervisor.send_system_message.call_args[0][0]) is Supervise


def test_actor_of_raise_if_factory_or_behavior_is_not_provider():
    factory = ActorFactory(Mock(), Mock())
    with pytest.raises(ValueError):
        factory.actor_of()



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from actors.actor import Actor
from actors.internal.cell import Cell
from actors.internal.messages import Resume, Failure, Start, Restart
from actors.ref import InternalRef
from .mock_compat import Mock


@pytest.fixture
def actor():
    actor = Mock(spec=Actor)
    actor.__class__ = Mock()
    actor.__class__.supervisor_strategy.return_value = Resume
    return actor


@pytest.fixture
def cell(actor):
    cell = Cell(lambda: actor, Mock(), Mock(), Mock())
    cell.handle_system_message(Start)
    return cell


def test_supervisor_should_ask_actor_class_for_strategy(actor, cell):
    cell.handle_system_message(Failure(Mock(), Mock(), Mock()))
    assert not actor.supervisor_strategy.called
    assert actor.__class__.supervisor_strategy.called


def test_supervisor_should_direct_child_on_failure(actor, cell):
    actor.__class__.supervisor_strategy.return_value = Restart
    child = Mock(spec=InternalRef)
    cell.handle_system_message(Failure(child, ValueError(), ""))
    child.send_system_message.assert_called_once_with(Restart)




# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from actors.actor import Actor


def as_proxy(actor_ref):
    """
    Create a proxy wrapper from an existing reference. The actor MUST implement the
    :class:`ProxyActor` interface.

    :param actor_ref: The actor reference.
    :type actor_ref: :class:`ActorRef`
    """
    return ProxyRef(actor_ref)


Invoke = namedtuple('Invoke', ['name', 'args', 'kwargs'])


class ProxyRef(object):

    def __init__(self, underlying):
        self._underlying = underlying

    def __getattr__(self, name):
        return lambda *args, **kwargs: \
            self._underlying.tell(Invoke(name, args, kwargs))


class ProxyActor(Actor):
    def receive(self, message):
        getattr(self, message.name)(*message.args, **message.kwargs)



# -*- coding: utf-8 -*-
import logging
import time
import random
from actors import Actor, ActorSystem, Directive

logging.basicConfig(level=logging.DEBUG)


class Crash(Exception):
    pass


StartGreeting = object()


class Greeter(Actor):
    def post_restart(self):
        print("Greeter restarted")

    def receive(self, message):
        if random.randint(0, 3) == 0:
            raise Crash()
        print("Hello %s" % message)


class Supervisor(Actor):
    @staticmethod
    def supervisor_strategy(exception):
        try:
            raise exception
        except Crash:
            return Directive.Restart
        except:
            return Directive.Stop

    def __init__(self):
        self._greeter = None

    def receive(self, message):
        if message is StartGreeting:
            self._greeter = self.context.actor_of(Greeter)
        else:
            # Forward the message to greeter
            self._greeter.tell(message, self.context.sender)


system = ActorSystem()
supervisor = system.actor_of(Supervisor)
supervisor.tell(StartGreeting)

try:
    while True:
        supervisor.tell("world")
        time.sleep(0.5)
except KeyboardInterrupt:
    pass

system.terminate()



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from actors.actor import Actor
from actors.system import ActorSystem
from actors.utils.ask import ask
from ..mock_compat import Mock


def test_system():
    class Test(Actor):
        def receive(self, message):
            self.context.sender.tell(message)

    with ActorSystem() as system:
        actor = system.actor_of(Test)
        message = object()
        reply = ask(actor, message)
        assert reply.get(timeout=1) is message


def test_behaviour():
    with ActorSystem() as system:
        behaviour = Mock()
        actor = system.actor_of(behaviour=behaviour)
        message = object()
        actor.tell(message)
        time.sleep(1)
        behaviour.assert_called_once_with(message)


def test_self_stop():
    on_post_stop = Mock()

    class Test(Actor):
        def receive(self, message):
            self.context.stop()

        def post_stop(self):
            on_post_stop()

    with ActorSystem() as system:
        actor = system.actor_of(Test)
        actor.tell(object())
        time.sleep(1)
        assert on_post_stop.called



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Lock
from actors.internal.executor import Executor
from actors.future import Promise


class Dispatcher(object):
    throughput = 5

    def __init__(self, executor):
        self._executor = executor
        self._attach_lock = Lock()
        self._attached_count = 0
        self._terminated = Promise()

    def dispatch(self, message, mailbox):
        if not mailbox.is_closed():
            mailbox.enqueue(message)
            self.schedule_execution(mailbox)
        else:
            print("Failed to deliver message. mailbox closed")

    def dispatch_system(self, message, mailbox):
        if not mailbox.is_closed():
            mailbox.enqueue_system(message)
            self.schedule_execution(mailbox)
        else:
            print("Failed to deliver system message. mailbox closed")

    def attach(self, mailbox):
        with self._attach_lock:
            self._attached_count += 1
        assert not mailbox.is_closed()

    def detach(self, mailbox):
        assert mailbox.is_closed()
        with self._attach_lock:
            self._attached_count -= 1
            if self._attached_count == 0:
                self._terminated.complete(None)
                self._executor.shutdown()

    def schedule_execution(self, mailbox):
        if mailbox.is_closed() or mailbox.is_scheduled() or not mailbox.has_messages():
            return

        if mailbox.set_scheduled():
            self._executor.submit(mailbox.process_messages)


class PinnedDispatcher(Dispatcher):

    def __init__(self):
        super(PinnedDispatcher, self).__init__(Executor(1))



from setuptools import setup

install_requires = ['six']

long_description = ""
with open('README.rst') as f:
    long_description += f.read()

with open('HISTORY.rst') as f:
    long_description += f.read()


setup(
    name='actors',
    version='0.5.1b1',
    author='Thomas Amland',
    author_email='thomas.amland@googlemail.com',
    url='https://github.com/tamland/python-actors/',
    description='Lightweight actor framework with supervision',
    long_description=long_description,
    license='Apache License 2.0',
    packages=['actors'],
    install_requires=install_requires,
    keywords=['actors', 'reactive', 'concurrent', 'threading', 'message passing', 'asynchronous'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy ',
        'Topic :: Software Development :: Libraries',
    ])



from actors import Actor, ActorSystem


class Greeter(Actor):
    def receive(self, message):
        print("Hello %s" % message)


with ActorSystem() as system:
    greeter = system.actor_of(Greeter)
    greeter.tell("world")



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

Start = object()
Restart = object()
Resume = object()
Terminate = object()

Restart = "restart"
Terminate = "terminate"

Failure = namedtuple('Failure', ['ref', 'exception', 'traceback'])
Supervise = namedtuple('Supervise', ['ref'])
Terminated = namedtuple('Terminated', ['ref'])

DeadLetter = namedtuple('DeadLetter', ['message', 'sender', 'recipient'])



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from actors.internal.executor import Executor
from .mock_compat import Mock


def test_executor_ignores_errors_in_task():
    executor = Executor()

    task1 = Mock()
    task1.side_effect = KeyError
    task2 = Mock()

    executor.submit(task1)
    executor.submit(task2)
    executor.shutdown()
    executor.join()

    task1.assert_called_once_with()
    task2.assert_called_once_with()



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import actors.actor
from actors.internal.messages import Start, Supervise
from actors.ref import ActorRef, InternalRef


def _actor_from_behaviour(behaviour):
    class Wrapper(actors.actor.Actor):
        def receive(self, message):
            behaviour(message)
    return Wrapper


class ActorFactory(object):

    def __init__(self, system, supervisor):
        """
        :param system:
        :type system: :class:`ActorSystem`

        :param supervisor:
        :type supervisor: :class:`InternalRef`
        """
        self._system = system
        self._supervisor = supervisor

    def actor_of(self, cls=None, behaviour=None, dispatcher=None):
        if cls:
            factory = cls
        elif behaviour:
            factory = _actor_from_behaviour(behaviour)
        else:
            raise ValueError()

        if dispatcher is None:
            dispatcher = self._system._system_dispatcher

        from actors.internal.cell import Cell
        cell = Cell(factory, dispatcher=dispatcher, system=self._system,
                         parent=self._supervisor)

        internal_ref = InternalRef(cell)
        self._supervisor.send_system_message(Supervise(internal_ref))
        internal_ref.send_system_message(Start)

        return ActorRef(cell)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import actors.internal.factory
from actors.internal.messages import Terminate


class ActorContext(actors.internal.factory.ActorFactory):
    """
    Provides contextual information to actors. Only valid within the
    :meth:`Actor.receive` method.

    :ivar sender: The sender of the current message.
    :vartype sender: :class:`ActorRef`

    :ivar self_ref: An :class:`ActorRef` to the this actor.
    :vartype self_ref: :class:`ActorRef`

    :ivar parent: The parent of this actor.
    :vartype parent: :class:`ActorRef`

    :ivar system: The system  this actor belongs to.
    :vartype system: :class:`ActorSystem`
    """

    sender = None
    self_ref = None
    parent = None
    system = None

    def stop(self):
        self.self_ref.send_system_message(Terminate)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
from six.moves.queue import Queue, Empty
from collections import namedtuple


Failure = namedtuple('Failure', ['exc_info'])


class Timeout(Exception):
    pass


class Future(object):
    def __init__(self):
        self._result = Queue(maxsize=1)
        self._success_callback = None
        self._failure_callback = None

    def get(self, timeout=None):
        """
        Return value on success, or raise exception on failure.
        """
        result = None
        try:
            result = self._result.get(True, timeout=timeout)
        except Empty:
            raise Timeout()

        if isinstance(result, Failure):
            six.reraise(*result.exc_info)
        else:
            return result

    def on_success(self, callback):
        self._success_callback = callback

    def on_failure(self, callback):
        self._failure_callback = callback


class Promise(object):
    def __init__(self):
        self._future = Future()

    def complete(self, result):
        if isinstance(result, Failure):
            self.failure(result)
        else:
            self.success(result)

    def success(self, result):
        self._future._result.put(result)
        if self._future._success_callback:
            self._future._success_callback(self._future)

    def failure(self, result):
        self._future._result.put(Failure(result))
        if self._future._failure_callback:
            self._future._failure_callback(self._future)

    @property
    def future(self):
        return self._future



# -*- coding: utf-8 -*-
#
# Copyright 2015 Thomas Amland
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from actors.ref import ActorRef
from actors.utils.ask import ask, PromiseActorRef
from .mock_compat import Mock


def test_ask_forwards_message_with_promise_actor_as_sender():
    actor = Mock(spec=ActorRef)
    message = object()
    ask(actor, message)

    assert actor.tell.call_count == 1
    assert actor.tell.call_args[0][0] is message
    assert isinstance(actor.tell.call_args[0][1], PromiseActorRef)


def test_tell_completes_future():
    actor = PromiseActorRef()
    actor.promise = Mock()
    message = object()
    actor.tell(message)

    actor.promise.complete.assert_called_once_with(message)
